REQUESTS_TIMEOUT = (5, 5)
TELEGRAM_CAPTION_LENGTH_LIMIT = 200

# Commands
COMMAND_START = "start"
COMMAND_LIST_APPROVALS = ["list", "l"]
COMMAND_APPROVE = ["approve", "a"]
COMMAND_REJECT = ["reject", "r"]
COMMAND_DELETE = ["delete", "d"]
COMMAND_CHATID = "chatid"

COMMAND_STATS = 'stats'

COMMAND_HELP = ['help', 'h']
COMMAND_VERSION = ['version', 'v']
COMMAND_CONFIG = ['config', 'c']

CANCEL_KEYBOARD_COMMAND = "/cancel_keyboard"

# inline keyboard button data
BUTTON_DATA_NOTHING = "_"
BUTTON_DATA_APPROVE = "a"
BUTTON_DATA_REJECT = "r"

# webserver

ENDPOINT_WEBHOOK = "/"

# api keys
KEY_ARCHIVED = "archived"
KEY_REJECTED = "rejected"
KEY_VOTES_REQUIRED = "votesRequired"
KEY_VOTES_RECEIVED = "votesReceived"
